/* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 749 $ $LastChangedDate: 2002-08-31 13:41:45 +0200 (Sat, 31 Aug 2002) $
 *
 * Runtime environment for gont.
 *
 * Assumes compiler that allows $ in identifiers, such as GCC.
 * (It is required to have Ksi, so this ain't so important)
 */

#include <gont/runtime.h>
#include <string.h>

/* Exceptions */

#define USE_TRY_STACK 1

struct try_block {
	void *longjmpbuf[5];
	void *value;
#if !USE_TRY_STACK
	struct try_block *chain;
#endif
};


static struct try_block *top;

#if USE_TRY_STACK
#define SET_SIZE 128
struct try_block_set {
	struct try_block_set *prev;
	struct try_block blocks[SET_SIZE];
};

static struct try_block *low, *high;

static struct try_block_set *cur, *cached;

struct try_block *gont_push_try_block(void)
{
	if (top == NULL) {
		if (low == NULL) {
			cur = gont_xmalloc0(sizeof(*cur));
			low = cur->blocks;
			high = cur->blocks + SET_SIZE - 1;
		}
		
		top = low;
	} else if (++top > high) {
		if (cached) {
			cached->prev = cur;
			cur = cached;
			cached = NULL;
		} else {
			struct try_block_set *p = gont_xmalloc0(sizeof(*cur));
			p->prev = cur;
			cur = p;
		}
		
		low = cur->blocks;
		high = cur->blocks + SET_SIZE - 1;

		top = low;
	}

	return top;
}

void *gont_pop_raised_value(void)
{
	void *v = top->value;

	memset(top, '\0', sizeof(*top));

	if (--top < low) {
		cached = cur;
		cur = cur->prev;
		low = cur->blocks;
		high = cur->blocks + SET_SIZE - 1;
		top = high;
	}

	return v;
}

#else
struct try_block *gont_push_try_block(void)
{
	struct try_block *t;

	t = gont_xmalloc(sizeof(*t));

	t->value = NULL;
	t->chain = top;
	
	return (top = t);
}

void *gont_pop_raised_value(void)
{
	void *v = top->value;

	top = top->chain;

	return v;
}
#endif

void gont_pop_try_block(void)
{
	gont_pop_raised_value();
}

void gont_raise(void *v)
{
	if (top == NULL)
		gont_oops("unhandled exception");

	top->value = v;
	__builtin_longjmp(top, 1);
	/* Dummy. */
	for (;;) {}
}

extern gont_ptroff Core$$Failure;
extern gont_ptroff Core$$Sys_error;
extern gont_ptroff Core$$Invalid_argument;

void gont_fail_with(const char *s)
{
	gont_raise(gont_make_union(Core$$Failure, gont_cstring(s)));
}

void gont_invalid_argument(const char *s)
{
	gont_raise(gont_make_union(Core$$Invalid_argument, gont_cstring(s)));
}

void gont_sys_error(gont_ptroff err)
{
	gont_raise(gont_make_union(Core$$Sys_error, 
				   gont_cstring(strerror(err))));
}
