/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 850 $ $LastChangedDate: 2002-09-01 22:13:49 +0200 (Sun, 01 Sep 2002) $
 *
 * Cg_expr - code generation for statements.
 */

open Ast;
open Util;
open Tyutil;

open Cg_util;
open Cg_types;
open Cg_globals;
open Cg_expr;

*(<Ksi::expr>option, Ksi::expr) 
match_and_init(cg_ctx c, Ksi::expr handle, pattern p, bool transport)
{
    switch p.a {
    case Void_pattern:
	return (None, nop);
	
    case As_pattern(_, v):
    case Var_pattern(v):
	_ ini;
	let s = get_binding(v.var_symbol) in {
	    _ val;
	    
	    if (transport) {
		let t = s.stype in
		val = Ksi::Convert(storage_type(t), 
				   t2s_conversion(t, handle));
	    } else {
		val = handle;
	    }

	    if (has_property(s, Is_guard_local) || 
		has_property(s, Is_let_variable))
		ini = vardef_init(c, s, val);
	    else
		ini = Ksi::Modify(local_ref(s), val);
        }

        switch p.a {
        case As_pattern(pat, _):
	    let (m, i) = match_and_init(c, handle, pat, transport) in
	    return (m, pair(ini, i));
        case _:
	    return (None, ini);
	}
	
    case Expr_pattern(e):
	Ksi::expr ex = compile_expr(c, e);
	if (transport)
	    let t = instanced_type_of(e) in
	    handle = Ksi::Convert(storage_type(t), 
				  t2s_conversion(t, handle));
	switch instanced_type_of(e) {
	case String:
	    return (Some(Ksi::Convert(Ksi::Bool, 
				      call("gont_streq", [ex, handle]))), nop);
	case _:
	    return (Some(bin(Ksi::Eq, ex, handle)), nop);
	}
	
    case Tuple_pattern(lst):
	Ksi::expr match = nop;
	Ksi::expr ini = nop;
	int pos = 0;
	
	void match_one(pattern pp)
	{
	    switch match_and_init(c, tuple_nth(handle, pos), pp, true) {
	    case (None, ex):
		ini = pair(ini, ex);

	    case (Some(x), ex):
		ini = pair(ini, ex);
		switch match {
		case Ksi::Nop:	match = x;
		case _:		match = bin(Ksi::Truth_and_if, match, x);
		}
	    }

	    pos++;
	}

	List::iter(match_one, lst);
	
	switch match {
	case Ksi::Nop:	return (None, ini);
	case _:		return (Some(match), ini);
	}

    case Struct_pattern(sp):
	Ksi::expr match = nop;
	Ksi::expr ini = nop;

	handle = Ksi::Convert(storage_type(sp.sp_type), handle);
	
	def match_one(s)
	{
	    let id = make_id(get_binding(s.sp_symbol).name) in
	    let ex = Ksi::Arrow(handle, id) in
	    switch match_and_init(c, ex, s.sp_pattern, true) {
	    case (None, ex):
		ini = pair(ini, ex);

	    case (Some(x), ex):
		ini = pair(ini, ex);
		switch match {
		case Ksi::Nop:	match = x;
		case _:		match = bin(Ksi::Truth_and_if, match, x);
		}
	    }
	}

	List::iter(match_one, sp.sp_entries);
	
	switch match {
	case Ksi::Nop:	return (None, ini);
	case _:		return (Some(match), ini);
	}

    case Union_pattern(u):
	let s = get_binding(u.np_symbol) in
	let tag = symbol_union_tag(s) in
	let cond = bin(Ksi::Eq, union_tag(handle), tag) in {
	    push_proto(s);
	    switch match_and_init(c, union_value(handle), u.np_pattern, true) {
	    case (None, ini):
		return (Some(cond), ini);
	    case (Some(ex), ini):
		return (Some(bin(Ksi::Truth_and_if, cond, ex)), ini);
	    }
	}
    }
}

Ksi::expr compile_matching(cg_ctx c, Ksi::expr handle, <case_clause>list clauses,
			   Ksi::expr header, Ksi::expr all_failed)
{
    Ksi::id block_id = make_id(genname("match_block"));
    Ksi::expr trailer = Ksi::Exit_block(block_id);
    
    *(Ksi::expr, /* cannot_fail */ bool) compile_case_clause(case_clause cc)
    {
	bool cannot_fail = false;
	Ksi::expr test_alternatives = nop;

	void compile_one_case(*(pattern, expr) oc)
	{
	    let (pat, cond) = oc in {
		if (cannot_fail) {
		    if (flags.w_pattern)
			warnl(pat.loc, "matching code compiled out");
		    return;
		}
		
		Ksi::expr test;
		
		let (match, ini) = match_and_init(c, handle, pat, false) in {
		    switch (match, cond.a.raw) {
		    case (None, Bool_const(true)):
			cannot_fail = true;
			test = Ksi::Bind(pair(ini, true_const));
			
		    case (Some(x), _):
			let ccond = pair(ini, compile_expr(c, cond)) in
			test = Ksi::Bind(bin(Ksi::Truth_and_if, x, ccond));
			
		    case (None, _):
			let ccond = compile_expr(c, cond) in
			test = Ksi::Bind(pair(ini, ccond));
		    }
		}

		switch test_alternatives {
		case Ksi::Nop:
		    test_alternatives = test;
		    
		case x:
		    test_alternatives = Ksi::Binary(Ksi::Truth_or_if, x, test);
		}
	    }
	}

	List::iter(compile_one_case, cc.cases);

	let body = compile_stmt(c, cc.body) in
	let all = Ksi::Compound(List::cons3(header, body, trailer)) in
	let vardefs = Ksi::Compound(List::map(partial(vardef, c), cc.patvars)) in
	let cond = Ksi::Cond(test_alternatives, all, nop) in
	return (Ksi::Bind(pair(vardefs, cond)), cannot_fail);
    }

    <Ksi::expr>list compile_clauses(<case_clause>list cl)
    {
	switch cl {
	case Cons(cc, tl):
	    let (ex, cannot_fail) = compile_case_clause(cc) in
	    if (cannot_fail) {
		if (flags.w_pattern) {
		    switch tl {
		    case Cons(hd, _):
			warnl(hd.body.loc, "matching code compiled out");
		    case Nil:
			skip;
		    }
		}
		
		return List::cons1(ex);
	    } else {
		return Cons(ex, compile_clauses(tl));
	    }
	    
	case Nil:
	    return List::cons1(all_failed);
	}
    }

    return Ksi::Block(block_id, Ksi::Compound(compile_clauses(clauses)));
}

Ksi::expr compile_switch(cg_ctx c, expr e, <case_clause>list clauses)
{
    let sw_id = make_id("sw$") in
    let vi = make_vardef(storage_type(type_of(e)), sw_id, compile_expr(c, e)) in
    let fail = call("gont_match_failure", []) in
    let matching = compile_matching(c, Ksi::Ref(sw_id), clauses, nop, fail) in
    return Ksi::Bind(pair(vi, matching));
}

Ksi::expr compile_let(cg_ctx c, <*(pattern, expr)>list assigns, <stmt>option s)
{
    Ksi::expr conds = nop;
    Ksi::expr inis = nop;
    int cnt = 0;

    def get_id()
    {
	switch s {
	case Some(_):
	    return make_id("let$" + itoa(cnt++));
	case None:
	    return make_id(genname("flet"));
	}
    }
    
    void one_assign(*(pattern, expr) assign)
    {
	let (pat, ex) = assign in
	let n = get_id() in
	let v = make_vardef(storage_type(type_of(ex)), n, compile_expr(c, ex)) in
	switch match_and_init(c, Ksi::Ref(n), pat, false) {
	case (Some(cond), ini):
	    let maybe_fail = Ksi::Cond(cond, nop, 
				       call("gont_match_failure", [])) in
	    conds = pair(conds, pair(v, maybe_fail));
	    inis = pair(inis, ini);
	    
	case (None, ini):
	    conds = pair(conds, v);
	    inis = pair(inis, ini);
	}
    }

    List::iter(one_assign, assigns);

    let lt = pair(conds, inis) in
    switch s {
    case Some(st):
	return Ksi::Bind(pair(lt, compile_stmt(c, st)));
    case None:
	return lt;
    }
}

/*
    exn = NULL
    try
      if push_ctx == local
        try
          if push_ctx == local
            TRY_BODY
        finally
          exn = pop_ctx
	  if exn
            WITH_BODY
          if exn
            raise exn
    finally
      exn = pop_ctx
      FINALLY_BODY
      if exn
        raise exn
*/

Ksi::expr compile_try(cg_ctx c, try_record tr)
{
    Ksi::id exn_value_id = make_id("exn$");
    Ksi::expr exn_ref_expr = Ksi::Ref(exn_value_id);

    def if_exn_not_null(e)
      (Ksi::Cond(bin(Ksi::Ne, exn_ref_expr, null_value), e, nop));
    
    Ksi::expr try_finally(Ksi::expr body, Ksi::expr fin)
    {
	Ksi::expr if_local_return(Ksi::expr e)
	{
	    let jmp_buf = call("gont_push_try_block", []) in
	    let jmp = Ksi::Built_in(make_id("setjmp"), [jmp_buf]) in
	    return Ksi::Cond(bin(Ksi::Eq, jmp, zero_const), e, nop);
	}
	
	let pop = Ksi::Modify(exn_ref_expr,
			      call("gont_pop_raised_value", [])) in
	let maybe_reraise = 
	    if_exn_not_null(call("gont_raise", [exn_ref_expr])) in
	return Ksi::Try_finally(if_local_return(body), 
			        pair(pair(pop, fin), maybe_reraise));
    }

    Ksi::expr with_matching(<case_clause>list clauses)
    {
	let kill_exn = Ksi::Modify(exn_ref_expr, null_value) in
	let name = make_id("exn_copy$") in
	let vi = make_vardef(storage_type(Exn), name, exn_ref_expr) in
	let m = compile_matching(c, Ksi::Ref(name), clauses, kill_exn, nop) in
	return if_exn_not_null(pair(vi, m));
    }
    
    let vi = make_vardef(storage_type(Exn), exn_value_id, null_value) in {
	Ksi::expr body = compile_stmt(c, tr.try_body);
	if (List::length(tr.try_with) != 0) {
	    body = try_finally(body, with_matching(tr.try_with));
	}

	switch tr.try_finally.a {
	case Skip:
	    skip;
	case _:
	    body = try_finally(body, compile_stmt(c, tr.try_finally));
	}

	return Ksi::Bind(pair(vi, body));
    }
}

Ksi::expr compile_nl_stmt(cg_ctx c, nl_stmt raw)
{
    switch raw {
    case Skip:
	return nop;
    
    case Expr(e):
	return Ksi::Convert(Ksi::Void, compile_expr(c, e));
    
    case Return(e):
	switch instanced_type_of(e) {
	case Void:
	    switch e.a.raw {
	    case Void_const:
		return Ksi::Return(nop);
	    case _:
		return pair(compile_expr(c, e), Ksi::Return(nop));
	    }
	case _:
	    return Ksi::Return(s2t_conversion(type_of(e), compile_expr(c, e)));
	}

    case Vardef(v):
	return vardef(c, get_binding(v.var_symbol));
	    
    case Block(lst):
	let body = List::map(partial(compile_stmt, c), lst) in
	return Ksi::Bind(Ksi::Compound(body));

    case Pair(s1, s2):
	return pair(compile_stmt(c, s1), compile_stmt(c, s2));
    
    case Switch(e, clauses):
	return compile_switch(c, e, clauses);
	    
    case Flying_let(assigns):
	return compile_let(c, assigns, None);
	
    case Let(assigns, s):
	return compile_let(c, assigns, Some(s));
    
    case If(e, s1, s2):
	return Ksi::Cond(compile_expr(c, e),
			 compile_stmt(c, s1),
			 compile_stmt(c, s2));
			 
    case Loop(l):
	let inner_loop = make_id("inner_loop$") in
	let brk = make_id("break$" + simple_name(l.ll_symbol)) in
	let cont = make_id("continue$" + simple_name(l.ll_symbol)) in
	let body = compile_stmt(c, l.ll_body) in
	let inc = compile_stmt(c, l.ll_inc) in
	let loop = Ksi::Loop(pair(Ksi::Block(cont, body), inc)) in
	return Ksi::Block(inner_loop, Ksi::Block(brk, loop));
    
    case Break(b):
	string prefix;
	switch b.break_kind {
	case Break_kind:	    prefix = "break$";
	case Continue_kind:	    prefix = "continue$";
	}
	return Ksi::Exit_block(make_id(prefix + simple_name(b.break_symbol)));
	    
    case Break_if_not(e):
	return Ksi::Cond(compile_expr(c, e), 
			 nop, 
			 Ksi::Exit_block(make_id("inner_loop$")));
    
    case Open(_):
	return nop;
    
    case Try(tr):
	return compile_try(c, tr);
    
    case Raise(e):
	return call("gont_raise", [compile_expr(c, e)]);
    }
}

Ksi::expr compile_stmt(cg_ctx c, stmt s)
{
    return wfl_locate(c, compile_nl_stmt, s);
}

// vim:sts=4:sw=4
