/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 841 $ $LastChangedDate: 2002-09-01 17:22:49 +0200 (Sun, 01 Sep 2002) $
 *
 * Cg_types - Gont-type -> Ksi-type mappings.
 */

open Ast;
open Util;
open Tyutil;
open Cg_util;

<Ksi::id>Hashtbl::t mangled_types = Hashtbl::create(203);
int mangled_id = 1;
int mangled_struct_id = 1;

// really initialized by push_basic_types()
Ksi::ty gt_int = Ksi::Void;
Ksi::ty gt_void = Ksi::Void;
Ksi::ty gt_exn = Ksi::Void;
Ksi::ty gt_string = Ksi::Void;
Ksi::ty gt_bool = Ksi::Void;
Ksi::ty gt_float = Ksi::Void;
Ksi::ty gt_tbool = Ksi::Void;
Ksi::ty gt_tfloat = Ksi::Void;
Ksi::ty gt_tvoid = Ksi::Void;

Ksi::ty flat_type(ty t, bool storage)
{
    Ksi::ty push_struct(<<sf_symbol>field>list lst)
    {
	_ make_field(<sf_symbol>field f)
	{
	    let s = get_binding(f.fld_symbol) in
	    return (flat_type(f.fld_type, true), make_id(s.name));
	}

	def inst(f) 
	    ({fld_type = canonical_type(f.fld_type),
	      fld_symbol = f.fld_symbol,
	      fld_loc = f.fld_loc});

	let m = "s:" + mangle_type(Struct(List::map(inst, lst))) in
	if (Hashtbl::mem(mangled_types, m)) {
	    return Ksi::Ptr(Ksi::Struct(Hashtbl::find(mangled_types, m)));
	} else {
	    Ksi::id n = make_id("s" + itoa(mangled_struct_id++));
	    Hashtbl::add(mangled_types, m, n);
	    let c = List::map(make_field, lst) in {
		output_decl(Ksi::Top_comment("mangled = " + m));
		output_decl(Ksi::Struct_def(Nil, n, c));
		return Ksi::Ptr(Ksi::Struct(n));
	    }
	}
    }
    
    switch type_instance(t) {
    case Int:			return ptroff;
    case String:		return gont_string;
    case Tuple(_):		return tuple_ptr;
    case Func(_):		return gont_lambda_expr;
    case Tyvar(_):		return poly;
    case X_tyvar(_):		return poly;
    case Struct(lst):		return push_struct(lst);
    case Opt_struct(lst):	return push_struct(lst);
    case Union(_):		return gont_union;
    case Abstract_type(_):	return ptroff;
    case Exn:			return gont_union;
    case Named_type(_): 
	return ice("Codegen::flat_type(): Named_type");
    case Bool:
	if (storage)    return Ksi::Bool;
	else		return ptroff;
    case Float:	
	if (storage)    return c_double;
	else		return c_double_ptr;
    case Void:
	// It's not passed this way when returned from function. We have
	// to special case it then.
	return ptroff;
    }
}

Ksi::expr size_of_struct(ty t)
{
    switch flat_type(t, true) {
    case Ksi::Ptr(Ksi::Struct(s)):
	return Ksi::Size_of(Ksi::Struct(s));
    case _:
	return ice("Cg_types::size_of_struct(): flat_type went crazy");
    } 
}

void push_basic_types()
{
    Ksi::ty ini(string name, Ksi::ty repr)
    {
	output_decl(Ksi::Ty_alias(Nil, make_id(name), repr));
	return Ksi::Tyref(make_id(name));
    }
    
    void struct2(string name, Ksi::ty t1, string f1, Ksi::ty t2, string f2)
    {
	output_decl(Ksi::Struct_def(Nil, make_id(name), 
				   List::cons2((t1, make_id(f1)), 
					       (t2, make_id(f2)))));
    }
    
    void proto0(Ksi::ty ret, string name)
    {
	output_decl(Ksi::Func_proto(List::cons1(Ksi::Extern),
				   ret, make_id(name), Nil));
    }
    
    void proto1(Ksi::ty ret, string name, Ksi::ty arg)
    {
	output_decl(Ksi::Func_proto(List::cons1(Ksi::Extern),
				   ret, make_id(name), List::cons1(arg)));
    }
    
    void proto2(Ksi::ty ret, string name, Ksi::ty arg1, Ksi::ty arg2)
    {
	output_decl(Ksi::Func_proto(List::cons1(Ksi::Extern),
				   ret, make_id(name), List::cons2(arg1, arg2)));
    }
    
    struct2("gont_string", 
	    ptroff, "size",
	    Ksi::Uint(8), "first_char");
    
    struct2("gont_lambda_expr", 
	    poly, "func_ptr",
	    poly, "closure");
	    
    struct2("gont_union", 
	    ptroff, "tag",
	    poly, "value");
	    
    gt_int = ini("g_int", flat_type(Int, true));
    gt_void = ini("g_void", flat_type(Void, true));
    gt_exn = ini("g_exn", flat_type(Exn, true));
    gt_string = ini("g_string", flat_type(String, true));
    gt_bool = ini("g_bool", flat_type(Bool, true));
    gt_float = ini("g_float", flat_type(Float, true));

    gt_tbool = ini("g_bool_t", flat_type(Bool, false));
    gt_tfloat = ini("g_float_t", flat_type(Float, false));
    gt_tvoid = ini("g_void_t", flat_type(Void, false));
    
    // gont_str{join,eq,cmp},
    // gont_make_{union,float,string,lambda_expr}
    // gont_match_failure, gont_raise, gont_push_try_block, 
    // gont_pop_raised_value

    proto2(gont_string, "gont_strjoin", gont_string, gont_string);
    proto2(ptroff, "gont_strcmp", gont_string, gont_string);
    proto2(ptroff, "gont_streq", gont_string, gont_string);

    proto2(gont_union, "gont_make_union", ptroff, poly);
    proto1(gt_tfloat, "gont_make_float", gt_float);
    proto2(gont_string, "gont_make_string", ptroff, Ksi::Ptr(Ksi::Uint(8)));
    proto2(gont_lambda_expr, "gont_make_lambda", poly, poly);
    proto2(tuple_ptr, "gont_make_tuple", ptroff, Ksi::Ellipsis);
    
    // TODO: set them no_return
    proto0(ksi_void, "gont_match_failure");
    proto1(ksi_void, "gont_raise", gt_exn);
    proto0(ksi_void, "gont_null_access");
    
    proto1(tuple_ptr, "gont_xmalloc", Ksi::Macro_type(make_id("uptr_off")));
    
    proto0(Ksi::Ptr(Ksi::Void), "gont_push_try_block");
    proto0(gt_exn, "gont_pop_raised_value");
}

Ksi::ty ksi_type(ty x, bool storage)
{
    Ksi::ty real_type(ty t)
    {
	if (is_type_instance(t))
	    return flat_type(strip_xt(t), storage);
	else
	    return ksi_type(type_instance_one(t), storage);
    }
    
    Ksi::id push_tyalias(ty t)
    {
	string m = mangle_type(t);

	if (storage)
	    m = "S:" + m;
	else
	    m = "T:" + m;
	
	
	if (Hashtbl::mem(mangled_types, m))
	    return Hashtbl::find(mangled_types, m);
	else {
	    Ksi::id i = make_id("t" + itoa(mangled_id++));
	    Hashtbl::add(mangled_types, m, i);
	    output_decl(Ksi::Ty_alias(List::cons1(Ksi::Export_name(m)), i,
				     real_type(t)));
	    return i;
	}
    }
    
    switch strip_xt(x) {
    case Int:	    return gt_int;
    case String:    return gt_string;
    case Exn:	    return gt_exn;
    case Bool:
	if (storage)	return gt_bool;
	else		return gt_tbool;
    case Float:	    
	if (storage)	return gt_float;
	else		return gt_tfloat;
    case Void:	    
	if (storage)	return gt_void;
	else		return gt_tvoid;

    case _:	    return Ksi::Tyref(push_tyalias(x));
    }

}

// Ksi type used to locally store value of type [t]
Ksi::ty storage_type(ty t)
{
    return ksi_type(t, true);
}

// Ksi type used to pass value of type [t] to function
Ksi::ty transport_type(ty t)
{
    return ksi_type(t, false);
}

bool need_t2s_conversion(ty t)
{
    switch type_instance(t) {
    case Bool:
    case Float:	    
	return true;

    case _:	    
	return false;
    }
}

bool need_s2t_conversion(ty t)
{
    switch type_instance(t) {
    case Bool:
    case Float:	    
	return true;
	
    case _:	    
	return false;
    }
}

Ksi::expr s2t_conversion(ty t, Ksi::expr e)
{
    switch type_instance(t) {
    case Bool:
	return Ksi::Convert(transport_type(Bool), e);
	
    case Float:	    
	return call("gont_make_float", [e]);
	
    case _:	    
	return e;
    }
}

Ksi::expr t2s_conversion(ty t, Ksi::expr e)
{
    switch type_instance(t) {
    case Bool:
	return Ksi::Binary(Ksi::Ne, Ksi::Convert(ptroff, e), zero_const);
	
    case Float:	    
	return Ksi::Indirect(Ksi::Convert(gt_tfloat, e));
	
    case _:	    
	return e;
    }
}

// vim:sts=4:sw=4
