/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * Copyright (c) 2002 Kamil Skalski.
 * Copyright (c) 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 858 $ $LastChangedDate: 2002-09-02 09:09:16 +0200 (Mon, 02 Sep 2002) $
 *
 * Patrange - check for pattern exhaustiveness and unused cases.
 *
 */
 
/* This has been completly rewritten, to do substraction not marges.  
   I still use ranges, as in old compiler, but algorithm works the other
   way around... how simple ;)  It's always hard to flip your ideas to
   the other side.  */

// someday-TODO: what to do:
// switch ex {
// case Foo: 
//  // ...
// case Baar: 
//  // ...
// }
// where Baar is spelling error? The problem isn't as big in switch, but
// it is in with { ... } matching. -Wcapitalization has been now implemented,
// that warns about capitalized pattern variables. We might want 
// -Wstrict-capitalization, that warns about non-capitalized constructors etc.
// -Wcapitalization should only warn about potentially dangerous things.

// someday-TODO: unused clause checks for with-matching

open Ast;
open Util;
open Tyutil;

union range {
    void Omega;	// match-all, `_' or `x'
    bool Bool_range;
    *(<ty>option, <<range>list>list) Tuple_range;
    *(var_symbol, <range>list) Union_range;
    expr Expr_range;
}

struct pr_ctx {
    <range>list pr_examples;
    bool pr_did_something;
}

pr_ctx make_ctx(<range>list examples)
{
    return {pr_examples = examples, pr_did_something = false};
}

string alternatives(<range>list l)
{
    return String::concat(" | ", List::map(string_of_range, l));
}

string string_of_range(range r)
{
    switch r {
    case Omega:		    return "_";
    case Bool_range(true):  return "true";
    case Bool_range(false): return "false";
    case Expr_range(e):	    return "some " + string_of_type(type_of(e));
	
    case Tuple_range(None, lst):
	return "(" + String::concat(", ", List::map(alternatives, lst)) + ")";
	       
    case Tuple_range(Some(t), lst):
	switch type_instance(t) {
	case Struct(flds):
	case Opt_struct(flds):
	    _ do_one(<sf_symbol>field f, _ ex)
	    {
		switch ex {
		case [Omega]:
		    return "";
		case [ex]:
		    return qualified_name(f.fld_symbol) + " = " + 
			   string_of_range(ex);
		case _:
		    return ice("Patrange::string_of_example(): non-1 (do_one)");
		}
	    }
	   
	    let all = List::map2(do_one, flds, lst) in
	    let filt = List::filter(fun (s) (s != ""), all) in
	    return "{" + String::concat(", ", filt) + "}";

	case _:
	    return ice("Patrange::string_of_example(): not struct");
	}
	       
    case Union_range(s, lst):
	return symbol_name(s) + "(" + alternatives(lst) + ")";
    }
}

range range_of_pattern(pattern p)
{
    range r;

    switch p.a {
    case Void_pattern:
	return Omega;
	
    case Var_pattern(v):
	return Omega;
    
    case As_pattern(pat, _):
	return range_of_pattern(pat);

    case Expr_pattern(e):
	switch e.a.raw {
	case Bool_const(x):
	    return Bool_range(x);
	case _:
	    return Expr_range(e);
	}

    case Tuple_pattern(lst):
	r = Tuple_range(None, List::map(fun (p) ([range_of_pattern(p)]), lst));

    case Struct_pattern(p):
	switch p.sp_entries {
	case Cons(se, _):
	    _ t = get_binding(se.sp_symbol).a.sf_tydef.stype;
	    switch type_instance(t) {
	    case Struct(c):
	    case Opt_struct(c):
		_ one_fld(<sf_symbol>field f)
		{
		    def is_our(s) 
		        (same_symbols(get_binding(s.sp_symbol), 
			              get_binding(f.fld_symbol)));
		    try {
			_ se = List::find(is_our, p.sp_entries);
			return [range_of_pattern(se.sp_pattern)];
		    } with {
		    case Not_found:
			return [Omega];
		    }
		}

		r = Tuple_range(Some(t), List::map(one_fld, c));

	    case _:
		return ice("Patrange::string_of_pattern(): not struct");
	    }
	case Nil:
	    return ice("Patrange::string_of_pattern(): empty");
	}
    
    case Union_pattern(np):
	let x = range_of_pattern(np.np_pattern) in
	r = Union_range(get_binding(np.np_symbol), [x]);
    }

    return r;
}

<var_symbol>list union_fields(var_symbol one_field)
{
    var_symbol get_fld_name(<var_symbol>field f)
    {
	return get_binding(f.fld_symbol);
    }
    
    switch union_of(one_field).stype {
    case Union(x):
	return List::map(get_fld_name, x);
    case _:
	return ice("Patrange::union_fields(): bad parent union");
    }
}

pr_ctx substract(pr_ctx ctx, range pat)
{
    pr_ctx ret = {pr_examples = [], pr_did_something = ctx.pr_did_something};
    
    def add_ex(ex)
    {
	ret.pr_examples = Cons(ex, ret.pr_examples);
    }

    def mark()
    {
	ret.pr_did_something = true;
    }
    
    pr_ctx list_substract(<range>list exs, <range>list pats)
    {
	return List::fold_left(substract, make_ctx(exs), pats);
    }
    
    def sub_from_one(ex)
    {
//dbgl("sub " + string_of_range(pat) + " from " + string_of_range(ex));
	switch (pat, ex) {
	case (Bool_range(x), Bool_range(y)) if (x == y):
	case (Omega, _):
	    mark();
	    
	case (Bool_range(x), Omega):
	    add_ex(Bool_range(!x));
	    mark();
	
	case (Bool_range(_), _):
	    add_ex(ex);
	
	case (Expr_range(_), _):
	    add_ex(pat);
	    mark();
	
	case (Union_range(s, _), Omega):
	    // gotta split union
	    let exs = List::map(fun (s) (Union_range(s, [Omega])), 
				union_fields(s)) in
	    List::iter(sub_from_one, exs);
    
	case (Union_range(sp, rp), Union_range(se, re)) 
					    if (same_symbols(sp, se)):
	    let uc = list_substract(re, rp);
	    if (uc.pr_did_something)
		mark();
	    switch uc.pr_examples {
	    case Nil:
		skip;
	    case _:
		add_ex(Union_range(se, uc.pr_examples));
	    }

	case (Union_range(_, _), _):
	    add_ex(ex);
	
	case (Tuple_range(t, lst), Omega):
	    // gotta split tuple
	    let flds = List::map(fun (_) ([Omega]), lst) in
	    sub_from_one(Tuple_range(t, flds));
	
	case (Tuple_range(tp, lp), Tuple_range(te, le)):
	    _ did = false;
	    
	    def one_fld(ex_before, ex_after, pat_lst)
	    {
		switch (ex_after, pat_lst) {
		case (Cons(ex_hd, ex_tl), Cons(pat_hd, pat_tl)):
		    one_fld(Cons(ex_hd, ex_before), ex_tl, pat_tl);

		    let tc = list_substract(ex_hd, pat_hd);
		    
		    if (tc.pr_did_something) {
			did = true;
		    
			switch tc.pr_examples {
			case Cons(_, _):
			    let before = List::rev(ex_before) in
			    let after = Cons(tc.pr_examples, ex_tl) in
			    let lst = List::concat(before, after) in
			    add_ex(Tuple_range(tp, lst));
			case Nil:
			    skip;
			}
		    }
			
		case (Nil, Nil):
		    skip;
		case _:
		    ice("Patrange::substract(): different tuple lengths");
		}
	    }

	    one_fld([], le, lp);
	    
	    if (did)
		mark();
	    else
		add_ex(ex);
	    
	case (Tuple_range(_), _):
	    ice("Patrange::substract(): tuple and non-tuple");
	}
    }

    List::iter(sub_from_one, ctx.pr_examples);

    return ret;
}

void check_exhaustiveness(<*(pattern, bool)>list pats)
{
    pr_ctx c = make_ctx([Omega]);

    if (!flags.w_pattern)
	return;
    
    def check_one(pp)
    {
	let (pat, can_fail) = pp;
	let r = range_of_pattern(pat);
	let pc = substract(c, r);

//dbgl(alternatives(c.pr_examples) +" -> "+alternatives(pc.pr_examples));

	if (!pc.pr_did_something)
	    warnl(pat.loc, "this match clause is unused");
	
	if (!can_fail)
	    c = pc;
    }

    List::iter(check_one, pats);

    switch c.pr_examples {
    case Nil:
	skip;	// ok
    case x:
	warn("this pattern matching is not exhaustive");
	warn("example of values that arn't matched: " + alternatives(x));
    }
}

// vim:sts=4:sw=4
