/* 
 * Copyright (c) 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 836 $ $LastChangedDate: 2002-09-01 14:52:17 +0200 (Sun, 01 Sep 2002) $
 *
 * Runtime environment for Gont, Core module implementation.
 */

open Core;

/* Stdout output. */
void print_int(int i)
{
    print_string(itoa(i)); 
}

void print_float(float f)
{
    print_string(ftoa(f)); 
}

void print_newline()
{
    print_string("\n");
}

void print_endline(string s)
{
    print_string(s);
    print_newline();
}

/* Stderr output. */
void prerr_int(int i)
{
    prerr_string(itoa(i)); 
}

void prerr_float(float f)
{
    prerr_string(ftoa(f)); 
}

void prerr_newline()
{
    prerr_string("\n");
}

void prerr_endline(string s)
{
    prerr_string(s);
    prerr_newline();
}

/* Input. */
int read_int()
{
    return atoi(read_string());
}

float read_float()
{
    return atof(read_string());
}

/* Pair opertations. */
'a fst(*('a, 'b) p)
{
    let (a, _) = p in
    return a;
}

'b snd(*('a, 'b) p)
{
    let (_, b) = p in
    return b;
}

// Warning about ,,unused global symbol Core::__dummy_disaster'' and 
// ,,dead code'' here is normal
void __dummy_disaster()
{
    // make sure these are linked (other are referenced elsewhere in
    // this module)
    raise Null_access;
    raise Match_failure;
    raise Not_found;
    raise End_of_file;
    raise Sys_error("");
    raise Invalid_argument("");
    raise Failure("");
}

/* Character type. */
typedef int char;

char char_of_int(int x)
{
    if (x < 0 || x > 255)
        raise Invalid_argument("char_of_int");
    return x;
}

int int_of_char(char x)
{
    return x;
}

/* String <-> bool conversion.  */
string string_of_bool(bool b)
{
    switch b {
    case true:
	return "true";
    case false:
	return "false";
    }
}

bool bool_of_string(string s)
{
    switch s {
    case "true":
	return true;
    case "false":
	return false;
    case _:
	raise Invalid_argument("Core::bool_of_string");
    }
}


/* Ref.  */
'a ref_get(<'a>ref r)
{
    return r.ref_contents;
}

void ref_set(<'a>ref r, 'a a)
{
    r.ref_contents = a;
}

<'a>ref ref_make('a a)
{
    return {ref_contents = a};
}

/* Utilities.  */

void ignore('a)
{
}

*('a ('b)) partial(*('a ('c, 'b)) f, 'c c)
{
	return (fun 'a ('b b) (f(c, b)));
}

*('a ('b)) partial2(*('a ('c, 'd, 'b)) f, 'c c, 'd d)
{
	return (fun 'a ('b b) (f(c, d, b)));
}


// vim:sts=4:sw=4
