/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 791 $ $LastChangedDate: 2002-08-31 20:38:09 +0200 (Sat, 31 Aug 2002) $
 *
 * Standard library for Gont, Array module interface.
 */

/** Array module provides mutable arrays indexed with integers.  Access time 
    is constant.  */

/* Type of arrays holding values of type ['a].  */
type <'a>t;

/* Given an array [a] and index in it [i] return value stored at this index.
   Raise [Invalid_argument] if [i] is outside [0] -- [Array::length(a)-1].  
   It is constant time operation.  */
'a get(<'a>t a, int i) = "Array__get";

/* Store [v] in [a] at position [i]. Raise [Invalid_argument] if [i] is 
   outside [0] -- [Array::length(a)-1].  It is constant time operation.  */
void set(<'a>t a, int i, 'a v) = "Array__set";

/* Return number of elements in array [a].  It is constant time operation.  */
int length(<'a>t a) = "Array__length";

/* Return fresh array of size [n], with all elements intialized to [v].  */
<'a>t make(int n, 'a v) = "Array__make";

/* Return fresh array of size [n] with 0th element initialized to [f(0)], 
   1st to [f(1)] and so on.  [f] is called with arguments incressing from
   [0] to [n-1].  */
<'a>t init(int n, *('a (int)) f) = "Array__init";

/* Run function [f] on each element of array [a], from [0] to 
   [Array::length(a)-1].  */
void iter(*(void ('a)) f, <'a>t a);

/* Return list containg all elements of array [a]. First element of list
   contains [Array::get(a, 0)], second [Array::get(a, 1)] and so on.  */
<'a>list to_list(<'a>t a);

/* Return fresh array containing elements of [l].  It holds, that
   [to_list(of_list(x)) == x].  */
<'a>t of_list(<'a>list l);

/* Resize given array [a] to [new_size] elements.  If array grows -- new
   elements are initialized to [v].  It takes [O(new_size)] time.  */
void resize(<'a>t a, int new_size, 'a v) = "Array__resize";

/* Return copy of array [src].  It takes [O(Array::length(src))] time.  */
<'a>t copy(<'a>t src) = "Array__copy";
