/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 809 $ $LastChangedDate: 2002-09-01 08:35:15 +0200 (Sun, 01 Sep 2002) $
 *
 * Standard library for Gont, String module implementation.
 */

open String;

string make(int n, char c)
{
	string s = create(n);
	if (int_of_char(c) != 0)
		fill(s, 0, length(s), c);
	return s;
}

string concat(string sep, <string>list lst)
{
	int sep_len = length(sep);
	
	int count_len(int len, <string>list l)
	{
		switch l {
		case Cons(x, xs):
			return count_len(len + sep_len + length(x), xs);
		case Nil:
			return len - sep_len;
		}
	}
	
	int len = count_len(0, lst);
	
	if (len <= 0)
		return "";
	
	string r = create(len);

	void blt(int pos, <string>list l)
	{
		switch l {
		case Cons(x, Nil):
			blit(x, 0, r, pos, length(x));
			
		case Cons(x, xs):
			blit(x, 0, r, pos, length(x));
			pos = pos + length(x);
			blit(sep, 0, r, pos, sep_len);
			pos = pos + sep_len;
			blt(pos, xs);
			
		case Nil:
			raise Failure("String::concat: impossible");
		}
	}

	blt(0, lst);

	return r;
}

void iter(*(void (char)) f, string s)
{
	int sz = length(s);
	for (int i = 0; i < sz; i++)
		f(get(s, i));
}

extern int String___index_from(string s, int pos, char c);
extern int String___rindex_from(string s, int pos, char c);

int index_from(string s, int pos, char c)
{
	int ret = String___index_from(s, pos, c);

	if (ret == -2)
		raise Invalid_argument("String::index_from");
	if (ret == -1)
		raise Not_found;
		
	return ret;
}

int rindex_from(string s, int pos, char c)
{
	int ret = String___rindex_from(s, pos, c);

	if (ret == -2)
		raise Invalid_argument("String::rindex_from");
	if (ret == -1)
		raise Not_found;
		
	return ret;
}

bool contains_from(string s, int pos, char c)
{
	int ret = String___index_from(s, pos, c);

	if (ret == -2)
		raise Invalid_argument("String::contains_from");
	
	return ret >= 0;
}

bool rcontains_from(string s, int pos, char c)
{
	int ret = String___rindex_from(s, pos, c);

	if (ret == -2)
		raise Invalid_argument("String::rcontains_from");
	
	return ret >= 0;
}

int index(string s, char c)
{
	return index_from(s, 0, c);
}

int rindex(string s, char c)
{
	return rindex_from(s, length(s) - 1, c);
}

bool contains(string s, char c)
{
	return contains_from(s, 0, c);
}

extern int String___index_str_from(string, int, string);

int index_str_from(string s, int pos, string c)
{
	int ret = String___index_str_from(s, pos, c);
	
	if (ret == -2)
		raise Invalid_argument("String::index_str_from");
	
	if (ret == -1)
		raise Not_found;
	
	return ret;
}

bool contains_str_from(string s, int pos, string c)
{
	int ret = String___index_str_from(s, pos, c);
	
	if (ret == -2)
		raise Invalid_argument("String::contains_str_from");
	
	return ret != -1;
}

int index_str(string s, string c)
{
	return index_str_from(s, 0, c);
}

bool contains_str(string s, string c)
{
	return contains_str_from(s, 0, c);
}

string capitalize(string s)
{
	s = String::copy(s);
	set(s, 0, uppercase_char(get(s, 0)));
	return s;
}

string uncapitalize(string s)
{
	s = String::copy(s);
	set(s, 0, lowercase_char(get(s, 0)));
	return s;
}

<string>list split(string ssep, string ss)
{
	<string>list split_rec(int pos, string sep, string s)
	{
		if (contains_str_from(s, pos, sep)) {
			int idx = index_str_from(s, pos, sep);
			if (idx == pos && length(sep) == 0) {
				idx++;
				if (idx > length(s))
					return Nil;
			}
			let x = sub(s, pos, idx - pos) in
			let rest = split_rec(idx + length(sep), sep, s) in
			return Cons(x, rest);
		} else {
			return List::cons1(sub(s, pos, length(s) - pos));
		}
	}

	return split_rec(0, ssep, ss);
}
