// type inference test
// $Rev: 782 $ $LastChangedDate: 2002-08-31 18:16:19 +0200 (Sat, 31 Aug 2002) $

/* Takes multiplication function, and one element. Returns power function,
 * with Time = O(log n). */
_ pow(_ mul, _ one)
{
	_ f(_ val, _ n) {
		if (n == 0)
			return one;
			
		if (n == 1)
			return val;
		
		_ res = f(val, n / 2);

		res = mul(res, res);
		
		if (n % 2 == 1)
			res = mul(res, val);

		return res;
	}

	return f;
}

*(int (int, int)) int_pow;

union expr {
	int Const;
        void Var;
        *(expr, expr) Add;
        *(expr, expr) Sub;
        *(expr, expr) Mul;
        *(expr, expr) Div;
        *(expr, int) Pow;
}

_ compute(_ e, _ val) {
        switch (e) {
        case Const(x):  return x;
        case Var:     return val;
        case Add(e1, e2): return compute(e1, val) + compute(e2, val);
        case Sub(e1, e2): return compute(e1, val) - compute(e2, val);
        case Mul(e1, e2): return compute(e1, val) * compute(e2, val);
        case Div(e1, e2): return compute(e1, val) / compute(e2, val);     
        case Pow(e1, n) : 
             int ex;
             ex = compute(e1, val);
             switch(n) {
             case 0: return 1;
             case 1: return ex;
             case x: 
                     if (n < 0) return 0;
                     else return ex * compute(Pow(e1, n-1), val);
             }

        }
}

void print_exp(_ e) {
	switch e {
        case Const(x): print_int(x);
	case Var():    print_string("x");
        case Mul(e1,e2): print_string("(");
			 print_exp(e1); 
                         print_string("*");
			 print_exp(e2);
                         print_string(")");   
        case Div(e1,e2): print_string("(");
			 print_exp(e1); 
                         print_string("/");
			 print_exp(e2);
                         print_string(")");   
        case Add(e1,e2): print_string("(");
			 print_exp(e1); 
                         print_string("+");
			 print_exp(e2);
                         print_string(")");   
        case Sub(e1,e2): print_string("(");
			 print_exp(e1); 
                         print_string("-");
			 print_exp(e2);
                         print_string(")");
	case Pow(ex, n): print_exp(ex);
                         print_string("^");
  			 print_int(n);   
        }
}

_ simplify(_ e) {
	_ ex;

 	switch e {
        case Add(e1, e2): ex = Add(simplify(e1), simplify(e2));
        case Sub(e1, e2): ex = Sub(simplify(e1), simplify(e2));
        case Mul(e1, e2): ex = Mul(simplify(e1), simplify(e2));
        case Div(e1, e2): ex = Div(simplify(e1), simplify(e2));
        case Pow(exx, n): ex =  Pow(simplify(exx),n);
        case x: ex = x;
        } 

	switch ex {
	case Mul(Const(0), _): return Const(0);
	case Mul(_, Const(0)): return Const(0);
	case Mul(Const(1), x): return x;
	case Mul(x, Const(1)): return x;
	case Add(x, Const(0)): return x;
	case Add(Const(0), x): return x;
	case Sub(x, Const(0)): return x;
	case Sub(Const(0), x): return x;
	case Mul(Const(k), Mul(Const(l), x)):
			return Mul(Const(k * l), x);
        case Add(Const(a), Const(b)): return Const(a+b);
        case Sub(Const(a), Const(b)): return Const(a-b);
        case Mul(Const(a), Const(b)): return Const(a*b);
        case Div(Const(a), Const(b)): return Const(a/b);
        case Pow(x, 1): return x;
        case Pow(x, 0): return Const(1);
        case Pow(Const(x), n): return Const(int_pow(x,n));
	case x: return x;
	}
}

_ diff(_ e) {
	_ r;

        switch e {
        case Const(_): r = Const(0);
        case Var():    r = Const(1);
        case Add(e1, e2): r = Add(diff(e1), diff(e2));
        case Sub(e1, e2): r = Sub(diff(e1), diff(e2));
        case Mul(e1, e2): 
               r = Add(Mul(diff(e1), e2), Mul(e1, diff(e2)));
        case Div(e1, e2):
               _ up = Sub(Mul(diff(e1), e2), Mul(e1, diff(e2)));
               _ down = Mul(e2, e2); 
               r = Div(up, down);
        case Pow(ex, n):
               r = Mul(Const(n), Mul(Pow(ex,n-1), diff(ex)));
        }

	return simplify(r);
}


section init
{
        /* f_power takes function f and int n, gives function f^n */ 

	_ f_power;

	f_power = pow(fun (f, g) 
		(fun (x) (f(g(x)))), 
			fun _ (_ x) (x));
        int_pow = pow(fun _ (int x, _ y) (x * y), 1);

        _ e = Sub(Add(Pow(Var, 6), Mul(Const(2), Pow(Var, 2))), Const(1)); 

        _ temp_ex;
        _ nth_diff;

	for (_ i = 0; i < 7; i = i + 1) {
	  	nth_diff = f_power(diff, i);
                temp_ex = nth_diff(e);
		temp_ex = simplify(temp_ex);
	        print_exp(temp_ex);
		print_string("\n");
	}

        print_int(compute(e, 2));
        print_string(", ");
        expr eprim = diff(e);
        print_int(compute(eprim, 2));
        print_string(", ");
        expr ebis = diff(eprim);
        print_int(compute(ebis, 2));
        print_string(".\n");
}

section init
{

	_ id = fun _ (_ x) (x);
	
	_ ble = fun (f) (fun (x) (f(x)));

	_ combine = 
		fun (f, g) 
			(fun (x) (f(g(x))));
	
	_ f = fun (_) (None);
	_ g = fun (x) {
		switch f(x) {
		case None:
			return "ble";
		case Some(_):
			return "foo";
		}
	};
}

/* 
 * expected-build-status: ok
 * run-test: yes
 * expected-output: ((x^6+(2*x^2))-1)
 * expected-output: ((6*x^5)+(4*x))
 * expected-output: ((30*x^4)+4)
 * expected-output: (120*x^3)
 * expected-output: (360*x^2)
 * expected-output: (720*x)
 * expected-output: 720
 * expected-output: 71, 200, 484.
 */
